/*------------------------------------------------------------------

SQ-150 Rev.A

 internal RC-Clock 8MHz
 FUSE  -FL11100100 

 (c)2008-09-23
 Motohiko Takeda

 2008-11-15 version 2
------------------------------------------------------------------ */
#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>

/*----------------------------------------------------------*/
#define	TRUE			1
#define	FALSE			0

// MIDIbZ[W
#define MIDI_NoteOff			0x80
#define MIDI_NoteOn				0x90
#define MIDI_PolykeyPressure	0xA0
#define MIDI_ControlChange		0xB0
#define MIDI_ProgramChange		0xC0
#define MIDI_ChannelPressure	0xD0
#define MIDI_PitchBend			0xE0

#define MIDI_StartSysEx			0xF0
#define MIDI_EndSysEx			0xF7

#define MIDI_TimingClock		0xF8
#define MIDI_Start				0xFA
#define MIDI_Continue			0xFB
#define MIDI_Stop				0xFC
#define MIDI_ActiveSensing		0xFE
#define MIDI_SystemReset		0xFF


#define F_CPU 8000000
#define UART_BAUD_RATE 31250

/* automatically calcuate baud register value */
#define UART_BAUD_SELECT (F_CPU/(UART_BAUD_RATE*16l)-1)

#define CLOCK_UP	0x01
#define CLOCK_DOWN	0x02
#define RUN_STOP 	0x04
#define INT_CLK 	0x08
#define ONE_SHOT 	0x10
#define START_INIT 	0x20


#define START_SW 	0x01		// Run / Srop switch 	(BITMAP)
#define INTERNAL_SW 0x02		// external/internal sw (BITMAP) 

#define	BUFSIZE		32		// Oobt@TCY
#define MAX_NOTE_CNT	8	//note buffer

void NoteOFF(void);
void NoteON(void);
void lfo(uint8_t speed);
void AD_convert();
unsigned char calc_LFO_speed(unsigned char value);
void sequencer(uint8_t step);
void do_clock_down(void);
void do_clock_up(void);
void do_gate_off(void);
void do_gate_on(void);

volatile uint8_t SEQ_STAT = 0x04 ;
volatile uint8_t AD_COMP = 0;
volatile uint8_t result = 70;

// Oobt@p
uint8_t	RxBuf[BUFSIZE];
volatile int	Ptr_buf_in, Ptr_buf_out; // |C^

uint8_t	RxByte;
uint8_t	PC;
uint8_t	SysEx;
uint8_t	Rch;
uint8_t	MByte[2];

uint8_t	PBBuf;
uint8_t	NoteCnt;
uint8_t	NoteBuf[MAX_NOTE_CNT];
uint8_t	PlayNoteBuf;
//-----------------------------------------------

// USART荞
ISR (USART_RX_vect)
{
	// Oobt@֕ۑ
	Ptr_buf_in++;
	Ptr_buf_in &= (BUFSIZE - 1);
	RxBuf[Ptr_buf_in] = UDR;
}

/*----------------------------------------------------------*/

ISR (TIMER1_OVF_vect)
{
	static uint8_t flag = 0;
	static uint8_t speed = 70;
	static uint8_t last_mode = 0 ;
	static uint8_t attack_cnt = 0 ;


	lfo(speed);	/*5..250*/

	if(SEQ_STAT & ONE_SHOT){			//makin' attack around here
		if (++attack_cnt > 16 ){
			SEQ_STAT &= ~ ONE_SHOT ; 	//0xef 
			attack_cnt = 0;
			if(SEQ_STAT & RUN_STOP){
				do_gate_on();
			}
		}
	}
	if (flag == 0 ){
		AD_convert();				//}jAe|̎o(`^O͂j

		if ( last_mode != PINA ){	//̓XCb`̎o
			if (PINA & ~START_SW  ) {		// port a 0 bit~(1 << START_SW ) 
				SEQ_STAT |= RUN_STOP;
				SEQ_STAT |= START_INIT;
			}else{
				SEQ_STAT &= ~RUN_STOP;	//0xfb 
				do_gate_off();
			}
			if (PINA & ~INTERNAL_SW ) {		//0xfd  port a 1 bit
				SEQ_STAT |= INT_CLK ;
			}else{
				SEQ_STAT &= ~INT_CLK;	// 0xf7
			}
		}
		last_mode = PINA;
	}
	if (flag == 128 ){			//ADRo[^[̃f[^e|f[^ɕϊăZbg
		speed = calc_LFO_speed(result );
	}
	flag ++ ;
}

/*----------------------------------------------------------*/

ISR (ANA_COMP_vect)
{
	result = TCNT0;    	//Load timer value
	TCCR0B = 0;
    PORTD &= ~(1<<PD6);	//	0xbf Start discharge
	AD_COMP = 0;
}

ISR (TIMER0_COMPA_vect)	//error trap
{
	TCCR0B = 0;			//Timer stop
	result = 0;			//Rescale A/D output
    PORTD &= ~(1<<PD6);	//	0xbf	Start discharge
	AD_COMP = 0;
}

void AD_convert()
{
	TCNT0 = 0;   		//Clear counter
	TCCR0B |= (1<<CS01);//Start timer0 with prescaling f/8
	PORTD |= (1<<PD6);	//Set converter charge
	AD_COMP = 1;
}
/*----------------------------------------------------------*/

unsigned char calc_LFO_speed(unsigned char value)
{
	unsigned int temp;		/*16bit */
	unsigned int speed;		/*16bit */
	uint8_t exp;

	if ( value > 110 ){
		return 250;
	}
	if ( value == 0 ){
		return 5;
	}

//	0 .. 110 ܂ł̒l277 .. 443܂łɃ}bsO
	temp = value * ((510-149)/110) + 149;
// vZ	 pow(2, temp/64 ) 搔0̎̏ȂBv  
	speed = 2 << ( (temp >> 6 )-1) ;
//	_ȉ͒ߎ
	exp = ((unsigned int) temp & 0x003F) + 64 ;  
//	Ə_ȉ	*/
	speed = (speed * (unsigned int)exp );	
//_ȉ؎̂ĂĐ
	speed >>= 6;
	return speed;
}

//-----------------------------------------------
// MIDIVXeA^CbZ[W

void MIDI_SystemRealtimeMessage(void)
{
	static uint8_t mck_cnt = 0; 
	switch(RxByte) {
		case MIDI_TimingClock:
			if ( !(SEQ_STAT & INT_CLK)  ){
				if (mck_cnt == 0 ){
					do_clock_up();
				}
				if (mck_cnt == 6 ){
					do_clock_down();
				}
				if (++ mck_cnt >= 12 ){
					mck_cnt = 0;
				}
			}
			break;
		case MIDI_Start:
			mck_cnt = 0;
			sequencer(-1);
			break;
		case MIDI_Continue:
			mck_cnt = 0;
			break;
		case MIDI_Stop:
			mck_cnt = 0;
			break;
		case MIDI_ActiveSensing:
			break;
		case MIDI_SystemReset:
			break;
	}
}

//-----------------------------------------------
// MIDIVXebZ[W

int MIDI_SystemMessage(void)
{
	if(SysEx){
		if(RxByte == MIDI_EndSysEx){
			SysEx = FALSE;
		}
	}
	else{
		if(RxByte < 0xF8) {
			if(RxByte > 0x7F){
				if(RxByte == MIDI_StartSysEx){
					SysEx = TRUE;
				}
				else{
					Rch = RxByte & 0x0F;
               	}
				PC = 0;
 			}
			else{
				MByte[PC & 0x01] = RxByte;
			}
			return TRUE;
		}
		else {
			MIDI_SystemRealtimeMessage();
		}
	}
	return FALSE;
}

//-----------------------------------------------
// MIDI`lbZ[W

void MIDI_ChannelMessage(void)
{
	switch(PC){
		case 0:
			switch(RxByte & 0xF0){
			case MIDI_NoteOff:
				PC = 2;
				break;
			case MIDI_NoteOn:
				PC = 4;
				break;
			case MIDI_PolykeyPressure:
				PC = 6;
				break;
			case MIDI_ProgramChange:
				PC = 8;
				break;
			case MIDI_ControlChange:
				PC = 10;
				break;
			case MIDI_ChannelPressure:
				PC = 12;
				break;
			case MIDI_PitchBend:
				PC = 14;
				break;
			} break;

 		case 2:			// m[gIt
			PC = 3;
			break;
        case 3:
			PC = 2;
			NoteOFF();
			break;

        case 4:			// m[gI
			PC = 5;
			break;
        case 5:
			PC = 4;
			if( MByte[1] == 0){
				NoteOFF();
			}
			else{
				NoteON();
			}
			break;

        case 6:			// |tHjbNL[vbV[ 
			PC = 7;
			break;
        case 7:
			break;

        case 8:			// vO`FW
 			break;

        case 10:		// Rg[`FW
			 PC = 11;
			 break;
        case 11:
			break;

		case 12:       // `lvbV[
			PC = 13;
			break;
		case 13:
			break;

		case 14:        // sb`xh
			PC = 15;
			break;
		case 15:
			break;
		default:
			break;
    }
}
//-----------------------------------------------
// m[gON
void NoteON(void)
{
	uint8_t i;
	uint8_t max = 0;

	// obt@ɓo^ς̃m[g͖
	for(i = 0; i < NoteCnt; i++) {
		if(NoteBuf[i] == MByte[0]) {
			return;
		}
	}

	// obt@ςH
	if(NoteCnt == MAX_NOTE_CNT) {
		// ʓ˂
		for(i = 0; i < (MAX_NOTE_CNT - 1); i++) {
			NoteBuf[i] = NoteBuf[i+1];
		}
		NoteBuf[MAX_NOTE_CNT - 1] = MByte[0];
	}
	else {
		NoteBuf[NoteCnt] = MByte[0];
		NoteCnt++;
	}

	// ō
	for(i = 0; i < NoteCnt; i++) {
		if(max < NoteBuf[i]) {
			max = NoteBuf[i];
		}
	}
	PlayNoteBuf = max;
	do_clock_up();
}

//-----------------------------------------------
// m[gOFF

void NoteOFF(void)
{
	uint8_t i;
	uint8_t max = 0;
	uint8_t flg = FALSE;

	// obt@ɓo^Ăꍇ͍폜
	for(i = 0; i < NoteCnt; i++) {
		if(flg) {
			NoteBuf[i-1] = NoteBuf[i];
		}
		if(NoteBuf[i] == MByte[0]) {
			flg = TRUE;
		}
	}
	if(flg) NoteCnt--;

	if(NoteCnt == 0) {
		// obt@J̏ꍇGATEIt
		do_clock_down();
		do_gate_off();
	}
	else {
		// ō
		for(i = 0; i < NoteCnt; i++) {
			if(max < NoteBuf[i]) {
				max = NoteBuf[i];
			}
		}
		PlayNoteBuf = max;
	}
}

void lfo(uint8_t speed)
{
	static int current_e = 0;
	static int current_level = 0 ;
	static int direction = 1;	/* 1:up, -1:down*/

	current_e += 5;		//q
	if (current_e>=  speed){
		current_e -= speed ;
		current_level += direction;
	}
	if (current_level > 255 ){	/*max should be 256*/
		direction = -1;
		current_e = 10;
		current_level -= 1 ;
		if (SEQ_STAT & INT_CLK  ){
			do_clock_down();
		}
	}
	if (current_level == 0){
		direction = +1;
		current_e = 10;
		current_level += 1;
		if (SEQ_STAT & INT_CLK  ){
			do_clock_up();
		}
	}
	OCR1A =  current_level<<2;
}

void do_clock_down(void)
{
	SEQ_STAT |=  CLOCK_DOWN;	// clock down
}

void do_clock_up(void)
{
	SEQ_STAT |= (CLOCK_UP | ONE_SHOT );	// clock up OneShot_Flag on
	do_gate_off();
}

void do_gate_off(void)		// put this port to inverter.
{
	PORTD &= ~(1<<PD4);		// 0xEF  
}

void do_gate_on(void)
{
	PORTD |= (1<<PD4) ;
}


void sequencer(uint8_t reset)
{
	static uint8_t SEQ_CNT = 0 ;

	if (reset ){
		SEQ_CNT = 0;
		PORTD &= ~(1<< PD5 ) ; 			//0xDF tempo indicator off 
		do_gate_off();
	}

	if (SEQ_CNT & (uint8_t)0x04) {
		PORTB |= (1<< PD2 );		//	LOWER LED on*/
	}else{
		PORTB &= ~( 1 << PB2 );	//0xFB UPPER LED on */
	}
	PORTB = ( PORTB & 0x0f) | ( 1 << ((SEQ_CNT & (uint8_t)0x03)+4 )) ;	/*LED sequence */

	PORTD = ( PORTD & 0xf1) | ( SEQ_CNT << 1) ;	/*405 drive */

	if (reset ){
		return;
	}
	if ( ++SEQ_CNT >= 8 ){
		SEQ_CNT = 0;
	}

}

void ioinit (void)
{
	/* TMR110-bit PWMœ */
	/* OC1Ao͐ݒ */
    DDRB |= (1<<PB3);

	TCCR1A |= (1<<COM1A1)|(1<<COM1A0)|(1<<WGM11)|(1<<WGM10);
	TCCR1B |= (1<<WGM12)|(1<<CS10);

	TIMSK |= (1<<TOIE1);

  	/* 10bit(1024)PWMl512 */
    OCR1A = 512;
    DDRB |= (1<<PB4)|(1<<PB5)|(1<<PB6)|(1<<PB7);	// for LED
	DDRB |= (1<<PB2);	// for LED Upper/Lower 
	

	/* PD o͐ݒ */
    DDRD |= (1<<PD1) | (1<<PD2) | (1<<PD3)|  (1<<PD4) | (1<<PD5);
	/*       4051-a  |  4051-b  |  4051-c |    gate   | tempo ind */

	/* PA ͐ݒ ivAbvj*/
    PORTA |= (1<<PA0) | (1<<PA1);


	/*MIDI Setup */
	UCSRA = 0x00;
	UCSRB |= (1 << RXCIE) | (1 << RXEN);
	UCSRC = 0x06;	// 8bit STOP1bit
	/* set baud rate */
	UBRRH = (UART_BAUD_SELECT) / 256; 
	UBRRL = (UART_BAUD_SELECT) % 256; 

	Ptr_buf_in = Ptr_buf_out = 0;

	/*AiORp[^[p̏-------------*/

    DIDR |= (1<<AIN1D)|(1<<AIN0D);	//fW^͋֎~WX^
									//convert_init()
	ACSR |= (1<<ACIE)|(1<<ACIS1)|(1<<ACIS0);	//and enable comparator interrupt
	TIMSK |= (1<<OCIE0A);			//Enable timer interrupt
	DDRD |= (1<<PD6);				// AD convert start port
    PORTD &= ~(1<<PD6);				//0xbf	Start discharge

	/*Clocl calibration 2008/11/15	*/
	OSCCAL = 90;

	sei ();
}

int main()
{
	ioinit();
	sequencer(-1);

	while(0 == 0){
		// Oobt@H
		if(Ptr_buf_in != Ptr_buf_out){	// Oobt@H
			Ptr_buf_out++;				// Oobt@PoCgo
			Ptr_buf_out &= (BUFSIZE - 1);
			RxByte = RxBuf[Ptr_buf_out];
			if(MIDI_SystemMessage()) {
				if(Rch == 0) {
					MIDI_ChannelMessage();
				}
			}
		}
		if ( SEQ_STAT & START_INIT ){
			SEQ_STAT &= ~START_INIT;	//0xdf
			sequencer(-1);
		}

		if ( SEQ_STAT & CLOCK_UP ){
			SEQ_STAT &= ~CLOCK_UP;			//0xfe;
			PORTD |= (1<< PD5 );	/*tempo indicator on */
			if ( (SEQ_STAT &  RUN_STOP) ||  !(SEQ_STAT &  INT_CLK ) ){	
				sequencer(0);
			}
		}
		if ( SEQ_STAT & CLOCK_DOWN ){
			SEQ_STAT &= ~CLOCK_DOWN;		//0xfd;
			PORTD &= ~(1<< PD5 ); 			/* 0xDF tempo indicator off */
		}	
	}

}

